import 'dart:convert';

import 'package:dio/dio.dart' as dio;
import 'package:http/browser_client.dart' as html;
import 'package:http/http.dart' as req;
import 'package:shared_preferences/shared_preferences.dart';
import 'enums.dart';
import 'exception.dart';
import 'base_client.dart';

BaseClient createClient(
        {required String endPoint,
        required bool selfSigned,
        req.Client? http}) =>
    BrowserClient(endPoint: endPoint, selfSigned: selfSigned, http: http);

class BrowserClient extends BaseClient {
  String endPoint;
  String type = 'unknown';
  Map<String, String>? headers;
  late Map<String, String> config;
  bool selfSigned;
  bool initialized = false;
  late req.Client http;
  late SharedPreferences _prefs;
  late dynamic cookieJar;

  BrowserClient(
      {this.endPoint = 'https://appwrite.io/v1',
      this.selfSigned = false,
      req.Client? http}) {
    // Platform is not supported in web so if web, set type to web automatically and skip Platform check
    this.http = http ?? req.Client();
    type = 'web';
    this.headers = {
      'content-type': 'application/json',
      'x-sdk-version': 'appwrite:flutter:0.6.4',
      'X-Appwrite-Response-Format': '0.8.0',
    };

    this.config = {};

    assert(endPoint.startsWith(RegExp("http://|https://")),
        "endPoint $endPoint must start with 'http'");
    init();
  }

  /// Your project ID
  BrowserClient setProject(value) {
    config['project'] = value;
    addHeader('X-Appwrite-Project', value);
    return this;
  }

  /// Your secret JSON Web Token
  BrowserClient setJWT(value) {
    config['jWT'] = value;
    addHeader('X-Appwrite-JWT', value);
    return this;
  }

  BrowserClient setLocale(value) {
    config['locale'] = value;
    addHeader('X-Appwrite-Locale', value);
    return this;
  }

  BrowserClient setSelfSigned({bool status = true}) {
    selfSigned = status;
    return this;
  }

  BrowserClient setEndpoint(String endPoint) {
    this.endPoint = endPoint;
    return this;
  }

  BrowserClient addHeader(String key, String value) {
    headers![key] = value;

    return this;
  }

  Future init() async {
    // if web set withCredentials true to make cookies work
    _prefs = await SharedPreferences.getInstance();
    addHeader('X-Fallback-Cookies', _prefs.getString('cookieFallback') ?? '');
    (http as html.BrowserClient).withCredentials = true;
    initialized = true;
  }

  Future<req.Response> call(
    HttpMethod method, {
    String path = '',
    Map<String, String> headers = const {},
    Map<String, dynamic> params = const {},
    dio.ResponseType? responseType,
  }) async {
    /* if (selfSigned && !kIsWeb) {
      // Allow self signed requests
      (http. httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate =
          (_HttpClient client) {
        client.badCertificateCallback =
            (X509Certificate cert, String host, int port) => true;
        return client;
      };
    } */

    if (!initialized) {
      await this.init();
    }

    params.removeWhere((key, value) => value == null);

    late req.Response res;
    try {
      headers.addAll(this.headers ?? {});
      var uri = Uri.parse('$endPoint$path');
      var request = req.Request(method.name(), Uri.parse(endPoint + path));
      request.headers.addAll(headers);
      if (headers['content-type'] == 'multipart/form-data') {
        request.bodyFields = Map<String, String>.from(params);
      } else if (method == HttpMethod.get) {
        params.keys.forEach((key) {
          if (params[key] is int || params[key] is double) {
            params[key] = params[key].toString();
          }
        });
        uri = Uri(
            fragment: uri.fragment,
            path: uri.path,
            host: uri.host,
            scheme: uri.scheme,
            queryParameters: params,
            port: uri.port);
      } else {
        request.body = jsonEncode(params);
      }
      final streamedResponse = await http.send(request);
      res = await req.Response.fromStream(streamedResponse);

      final cookieFallback = res.headers['x-fallback-cookies'];
      if (cookieFallback != null) {
        print(
            'Appwrite is using localStorage for session management. Increase your security by adding a custom domain as your API endpoint.');
        addHeader('X-Fallback-Cookies', cookieFallback);
        _prefs.setString('cookieFallback', cookieFallback);
      }
      return res;
    } on dio.DioError catch (e) {
      if (e.response == null) {
        throw AppwriteException(e.message);
      }
      if (responseType == dio.ResponseType.bytes) {
        if ((e.response!.headers.value('content-type') ?? '')
            .contains('application/json')) {
          final res = json.decode(utf8.decode(e.response!.data));
          throw AppwriteException(res['message'], res['code'], e.response);
        } else {
          throw AppwriteException(e.message);
        }
      }
      if ((e.response!.headers.value('content-type') ?? '')
          .contains('application/json')) {
        throw AppwriteException(e.response!.data['message'],
            e.response!.data['code'], e.response!.data);
      } else {
        throw AppwriteException(
            e.response!.data, e.response!.statusCode, e.response!.data);
      }
    } catch (e) {
      throw AppwriteException(e.toString());
    }
  }
}
