import 'dart:convert';
import 'dart:io';
import 'package:cookie_jar/cookie_jar.dart';
import 'package:device_info_plus/device_info_plus.dart';
import 'package:dio/dio.dart' as dio;
import 'package:http/http.dart' as req;
import 'package:package_info_plus/package_info_plus.dart';
import 'package:path_provider/path_provider.dart';
import 'base_client.dart';
import 'cookie_manager.dart';
import 'enums.dart';
import 'exception.dart';
import 'interceptor.dart';

BaseClient createClient(
        {required String endPoint,
        required bool selfSigned,
        req.Client? http}) =>
    IOClient(endPoint: endPoint, selfSigned: selfSigned, http: http);

class IOClient extends BaseClient {
  String endPoint;
  String type = 'unknown';
  Map<String, String>? headers;
  late Map<String, String> config;
  bool selfSigned;
  bool initialized = false;
  late req.Client http;
  late dynamic cookieJar;
  List<Interceptor> interceptors = [];

  IOClient(
      {this.endPoint = 'https://appwrite.io/v1',
      this.selfSigned = false,
      req.Client? http}) {
    // Platform is not supported in web so if web, set type to web automatically and skip Platform check
    this.http = http ?? req.Client();
    type = (Platform.isIOS) ? 'ios' : type;
    type = (Platform.isMacOS) ? 'macos' : type;
    type = (Platform.isAndroid) ? 'android' : type;
    type = (Platform.isLinux) ? 'linux' : type;
    type = (Platform.isWindows) ? 'windows' : type;
    type = (Platform.isFuchsia) ? 'fuchsia' : type;

    this.headers = {
      'content-type': 'application/json',
      'x-sdk-version': 'appwrite:flutter:0.6.4',
      'X-Appwrite-Response-Format': '0.8.0',
    };

    this.config = {};

    assert(endPoint.startsWith(RegExp("http://|https://")),
        "endPoint $endPoint must start with 'http'");
    init();
  }

  Future<Directory> _getCookiePath() async {
    final directory = await getApplicationDocumentsDirectory();
    final path = directory.path;
    final Directory dir = new Directory('$path/cookies');
    await dir.create();
    return dir;
  }

  /// Your project ID
  IOClient setProject(value) {
    config['project'] = value;
    addHeader('X-Appwrite-Project', value);
    return this;
  }

  /// Your secret JSON Web Token
  IOClient setJWT(value) {
    config['jWT'] = value;
    addHeader('X-Appwrite-JWT', value);
    return this;
  }

  IOClient setLocale(value) {
    config['locale'] = value;
    addHeader('X-Appwrite-Locale', value);
    return this;
  }

  IOClient setSelfSigned({bool status = true}) {
    selfSigned = status;
    return this;
  }

  IOClient setEndpoint(String endPoint) {
    this.endPoint = endPoint;
    return this;
  }

  IOClient addHeader(String key, String value) {
    headers![key] = value;

    return this;
  }

  Future init() async {
    // if web skip cookie implementation and origin header as those are automatically handled by browsers
    final Directory cookieDir = await _getCookiePath();
    cookieJar = new PersistCookieJar(storage: FileStorage(cookieDir.path));
    this.interceptors.add(CookieManager(cookieJar));
    PackageInfo packageInfo = await PackageInfo.fromPlatform();
    addHeader('Origin', 'appwrite-$type://${packageInfo.packageName}');

    //creating custom user agent
    DeviceInfoPlugin deviceInfoPlugin = DeviceInfoPlugin();
    var device = '';
    if (Platform.isAndroid) {
      final andinfo = await deviceInfoPlugin.androidInfo;
      device =
          '(Linux; U; Android ${andinfo.version.release}; ${andinfo.brand} ${andinfo.model})';
    }
    if (Platform.isIOS) {
      final iosinfo = await deviceInfoPlugin.iosInfo;
      device = '${iosinfo.utsname.machine} iOS/${iosinfo.systemVersion}';
    }
    if (Platform.isLinux) {
      final lininfo = await deviceInfoPlugin.linuxInfo;
      device = '(Linux; U; ${lininfo.id} ${lininfo.version})';
    }
    if (Platform.isWindows) {
      final wininfo = await deviceInfoPlugin.windowsInfo;
      device =
          '(Windows NT; ${wininfo.computerName})'; //can't seem to get much info here
    }
    if (Platform.isMacOS) {
      final macinfo = await deviceInfoPlugin.macOsInfo;
      device = '(Macintosh; ${macinfo.model})';
    }
    addHeader(
        'user-agent', '${packageInfo.appName}/${packageInfo.version} $device');

    // this.http.options.baseUrl = this.endPoint;
    // this.http.options.validateStatus = (status) => status! < 400;
    initialized = true;
  }

  Future<req.Request> _interceptRequest(req.Request request) async {
    final body = request.body;
    for (final i in interceptors) {
      if (i is Interceptor) {
        request = await i.onRequest(request);
      }
    }

    assert(
      body == request.body,
      'Interceptors should not transform the body of the request'
      'Use Request converter instead',
    );
    return request;
  }

  Future<req.Response> _interceptResponse(req.Response response) async {
    final body = response.body;
    for (final i in interceptors) {
      if (i is Interceptor) {
        response = await i.onResponse(response);
      }
    }

    assert(
      body == response.body,
      'Interceptors should not transform the body of the response',
    );
    return response;
  }

  Future<req.Response> call(
    HttpMethod method, {
    String path = '',
    Map<String, String> headers = const {},
    Map<String, dynamic> params = const {},
    dio.ResponseType? responseType,
  }) async {
    /* if (selfSigned && !kIsWeb) {
      // Allow self signed requests
      (http. httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate =
          (_HttpClient client) {
        client.badCertificateCallback =
            (X509Certificate cert, String host, int port) => true;
        return client;
      };
    } */

    if (!initialized) {
      await this.init();
    }

    params.removeWhere((key, value) => value == null);

    late req.Response res;
    try {
      headers.addAll(this.headers ?? {});
      var uri = Uri.parse('$endPoint$path');
      var request = req.Request(method.name(), Uri.parse(endPoint + path));
      request.headers.addAll(headers);
      if (headers['content-type'] == 'multipart/form-data') {
        request.bodyFields = Map<String, String>.from(params);
      } else if (method == HttpMethod.get) {
        params.keys.forEach((key) {
          if (params[key] is int || params[key] is double) {
            params[key] = params[key].toString();
          }
        });
        uri = Uri(
            fragment: uri.fragment,
            path: uri.path,
            host: uri.host,
            scheme: uri.scheme,
            queryParameters: params,
            port: uri.port);
      } else {
        request.body = jsonEncode(params);
      }
      request = await _interceptRequest(request);
      final streamedResponse = await http.send(request);
      res = await req.Response.fromStream(streamedResponse);
      res = await _interceptResponse(res);
      return res;
    } on dio.DioError catch (e) {
      if (e.response == null) {
        throw AppwriteException(e.message);
      }
      if ((e.response!.headers.value('content-type') ?? '')
          .contains('application/json')) {
        throw AppwriteException(e.response!.data['message'],
            e.response!.data['code'], e.response!.data);
      } else {
        throw AppwriteException(
            e.response!.data, e.response!.statusCode, e.response!.data);
      }
    } catch (e) {
      throw AppwriteException(e.toString());
    }
  }
}
